%----------------------------------------------%
%   Parser problem			       %
%   Parallel Prolog version		       %
%   Processor Farm Solution		       %
%   with Processor Manager		       %
%   with graphics			       %
%   written by Peter Kacsuk		       %
%----------------------------------------------%

%   calling parameters: N numbers of word_lists
%			D delay value
%			P number of processors

problem(N,D,P):-
   is_int(N), N > 0, N < 21,
   is_int(D), D > 0,
   max_processor(MP),
   is_int(P), P > 0, P <= MP,
   start(N,D,P).
problem(_,_,_):-
   write("Sorry, parameter error.").

start(N,D,P):-
   cpu_time(ST),
   init_windows,
   new(master(N,P,ST),master),
   start_workers(D,P).

start_workers(D,1):- !.
start_workers(D,N):-
   new(worker(D,N,0),worker(N),_,_,N),
   M is N - 1,
   start_workers(D,M).

%----------------------------------------------%
%		Master			       %
%----------------------------------------------%

master(N,P,ST):-
   create_master_window,
   delete_partition(free),
   create_free_processors(P),
   distribute_task(1,N),
   M is P - 1,
   collect_result(M),
   cpu_time(ET),
   T is ET - ST,
   display_time(T),
   terminate_system.

distribute_task(N,MAX):-
   N > MAX.
distribute_task(N,MAX):-
   find_clause(word_list(N,L)),
   search_for_free_processor(T), !,
   send(w(L,N),[worker(T)]),
   M is N + 1 ,
   distribute_task(M,MAX).
distribute_task(M,MAX):-
   wait_for_unb(res(L,R,N,T)),
   display_result(L,R,N),
   update_processor(T),
   distribute_task(M,MAX).

collect_result(M):-
   any_busy_processor(M),
   wait_for_unb(res(L,R,N,T)),
   display_result(L,R,N),
   update_processor(T),
   collect_result(M).
collect_result(_).

%-----------------------------%
% processor manager interface %
%-----------------------------%

create_free_processors(1).
create_free_processors(N):-
   add_clause(free(N)),
   M is N - 1,
   create_free_processors(M).

search_for_free_processor(T):-
   find_clause(free(T),I),
   delete_clause(free,I).

any_busy_processor(M):-
   not(clause_count(free,M)).

update_processor(T):-
   add_clause(free(T)).

%-----------------------------%
%    word list data base      %
%-----------------------------%

word_list(1,[the,man,eats,pork]).
word_list(2,[the,good,man]).
word_list(3,[the,man,eats]).
word_list(4,[man,drinks,wine]).
word_list(5,[the,man,eats,the,pork]).
word_list(6,[man,eats,pork]).
word_list(7,[man,eats,man]).
word_list(8,[the,man,eats,the,pork]).
word_list(9,[the,man,eats,the,boiled,pork]).
word_list(10,[man,drinks,pork]).
word_list(11,[man,eats,the,pork]).
word_list(12,[the,man,eats,the,beef]).
word_list(13,[the,pork,the,beef]).
word_list(14,[the,man,wine]).
word_list(15,[the,man,drinks,the,wine]).
word_list(16,[the,pork,drinks,the,dry,wine]).
word_list(17,[the,man,and,pork,drinks,wine]).
word_list(18,[man,drinks,the,wine]).
word_list(19,[man,drinks,beer]).
word_list(20,[the,man,and,the,pork,eats,wine]).

%----------------------------------------------%
%		Worker			       %
%----------------------------------------------%

worker(D,T,Cin):-
   set_value(delay_value,D),
   wait_for_unb(w(L,N)),
   write_to_string(ST,T),
   concat("aw",ST,AWid),
   concat("w",ST,Wid),
   concat("Worker",ST,Name),
   activate_window(AWid,Name,L),
   sentence(L,R,Cin,Cout),
   write_to_string(SC,Cout),
   deactivate_window(Wid,Name,SC),
   send(res(L,R,N,T),[master]),
   worker(D,T,Cout).

sentence(X,succ,Cin,Cout):-
   C is Cin + 1,
   set_value(counter,C),
   get_value(delay_value,D),
   delay(D),
   append(Y,Z,X), not(eq(Y,[])), not(eq(Z,[])),
   noun_phrase(Y),
   verb_phrase(Z),
   get_value(counter,Cout),
   !.
sentence(X,fail,_,Cout):-
   get_value(counter,Cout).

noun_phrase(X):-
   delay_counter,
   append(Y,Z,X), not(eq(Y,[])), not(eq(Z,[])),
   delay_counter, determiner(Y),
   delay_counter, noun(Z).

verb_phrase(X):-
   delay_counter,
   append(Y,Z,X), not(eq(Y,[])), not(eq(Z,[])),
   delay_counter, verb(Y),
   noun_phrase(Z).
verb_phrase(X):-
   delay_counter, verb(X).

determiner([the]).

noun([man]).
noun([pork]).
noun([wine]).

verb([eats]).
verb([drinks]).

append([],L,L).
append([X|L1],L2,[X|L3]):-
	append(L1,L2,L3).

delay_counter:-
   get_value(counter,Cin),
   C is Cin + 1,
   set_value(counter,C),
   get_value(delay_value,D),
   delay(D).

%----------------------------------------------%
%		Window handler		       %
%----------------------------------------------%

init_windows:-
   create_deactivated_windows,
   create_activated_windows,
   draw_worker_windows.

create_deactivated_windows:-
   create_window(2,3,3,35,628,w1),
   create_window(2,42,3,35,628,w2),
   create_window(20,3,3,35,628,w3),
   create_window(20,42,3,35,628,w4).

create_activated_windows:-
   create_window(2,3,3,35,608,aw1),
   create_window(2,42,3,35,592,aw2),
   create_window(20,3,3,35,560,aw3),
   create_window(20,42,3,35,544,aw4).

create_master_window:-
   set_value(line,1),
   create_window(7,17,11,46,590,res),
   open_window(res),
   write_window(res,0,12,"MASTER").

draw_worker_windows:-
   deactivate_window(w1,"Worker1","0"),
   deactivate_window(w2,"Worker2","0"),
   deactivate_window(w3,"Worker3","0"),
   deactivate_window(w4,"Worker4","0").

deactivate_window(Wid,Name,Count):-
   open_window(Wid),
   write_window(Wid,0,11,Name),
   write_window(Wid,0,2,Count).

activate_window(Wid,Name,L):-
   write_to_string(SL,L),
   open_window(Wid),
   write_window(Wid,0,11,Name),
   write_window(Wid,1,1,SL).

display_result(L,R,N):-
   get_value(line,LN),
   write_window(res,LN,1,"                                                "),
   write_to_string(SR,R),
   write_window(res,LN,1,SR),
   write_window(res,LN,5,":"),
   write_to_string(SN,N),
   write_window(res,LN,7,SN),
   write_to_string(SL,L),
   write_window(res,LN,10,SL),
   NL is (LN mod 9) + 1,
   set_value(line,NL).

display_time(T):-
   create_window(12,1,3,15,590,wt),
   open_window(wt),
   write_window(wt,0,7,"time"),
   write_to_string(ST,T),
   write_window(wt,1,1,ST).

delay(0):- !.
delay(N):-
   M is N - 1 ,
   delay(M).

